from .Expand import Expand
from .Crossover import CrossOver
from .GenerateSimilar import GenerateSimilar
from .Rephrase import Rephrase
from .Shorten import Shorten
from .AlterSentenceStructure import AlterSentenceStructure
from .ChangeStyle import ChangeStyle
from .InsertMeaninglessCharacters import InsertMeaninglessCharacters
from .MisspellSensitiveWords import MisspellSensitiveWords
from .Translation import Translation
from .historical_insight import HistoricalInsight
